@@ -17,6 +17,9 @@
 #ifndef __MOD_CHXJ_H__
 #define __MOD_CHXJ_H__
 
+#define CHXJ_TRUE    (1)
+#define CHXJ_FALSE   (0)
+
 #ifdef UNUSED
 #elif defined(__GNUC__)
 # define UNUSED(x) UNUSED_ ## x __attribute__((unused))
@@ -55,6 +58,8 @@
 
 #define STRCASEEQ(a,b,c,d) \
   ((((a) == *(d))|| ((b) == *(d))) && strcasecmp((c),(d)) == 0)
+#define STRNCASEEQ(a,b,c,d,e) \
+  ((((a) == *(d))|| ((b) == *(d))) && strncasecmp((c),(d),(e)) == 0)
 
 #include "qs_ignore_sp.h"
 #include "qs_log.h"
@@ -277,6 +282,11 @@ typedef enum {
 
 typedef struct mod_chxj_config mod_chxj_config;
 
+#if defined(USE_MYSQL_COOKIE)
+#  include "chxj_mysql.h"
+#elif defined(USE_MEMCACHE_COOKIE)
+#  include "chxj_memcache.h"
+#endif
 struct mod_chxj_config {
   int                   image;
 
@@ -297,6 +307,12 @@ struct mod_chxj_config {
 
   char*                 cookie_db_dir;
   long                  cookie_timeout;
+
+#if defined(USE_MYSQL_COOKIE)
+  mysql_t               mysql;
+#elif defined(USE_MEMCACHE_COOKIE)
+  memcache_t            memcache;
+#endif
 };
 
 
@@ -367,6 +383,9 @@ extern char* chxj_node_exchange(
 
 #define IMAGE_CACHE_LIMIT_FMT_LEN  (20)
 
+#if HAVE_MALLOC == 0
+extern void *rpl_malloc(size_t n);
+#endif
 
 #endif
 /*
